@@ -33,6 +33,7 @@
 #include <glade/glade.h>
 #include <gconf/gconf-client.h>
 #include <libwnck/libwnck.h>
+#include <libgnomevfs/gnome-vfs-mime-handlers.h>
 #include "eggtrayicon.h"
 
 #include "acme.h"
@@ -53,6 +54,11 @@ typedef struct {
 	guint dialog_timeout;
 	WnckScreen *screen;
 
+	/* Multihead stuff */
+	GdkDisplay *display;
+	GdkScreen *current_screen;
+	GList *screens;
+
 	/* Tray icon */
 	EggTrayIcon *tray_icon;
 	GtkTooltips *tray_icon_tooltip;
@@ -294,11 +300,12 @@ acme_image_set (Acme *acme, int icon)
 }
 
 static void
-ungrab_key (int key_code)
+ungrab_key_real (int key_code, GdkWindow *root)
 {
 	gdk_error_trap_push ();
 	XUngrabKey (GDK_DISPLAY (), key_code, AnyModifier,
-			GDK_ROOT_WINDOW ());
+			GDK_WINDOW_XID (root));
+
 	gdk_flush ();
 	if (gdk_error_trap_pop ()) {
 		char *error;
@@ -314,41 +321,56 @@ ungrab_key (int key_code)
 }
 
 static void
-grab_key (int key_code)
+ungrab_key (Acme *acme, int key_code)
+{
+	GList *l;
+
+	for (l = acme->screens; l != NULL; l = l->next)
+	{
+		GdkScreen *screen;
+
+		screen = (GdkScreen *) l->data;
+		ungrab_key_real (key_code, gdk_screen_get_root_window (screen));
+	}
+}
+
+
+static void
+grab_key_real (int key_code, GdkWindow *root)
 {
 	gdk_error_trap_push ();
 
 	XGrabKey (GDK_DISPLAY (), key_code,
 			0,
-			GDK_ROOT_WINDOW (), True,
+			GDK_WINDOW_XID (root), True,
 			GrabModeAsync, GrabModeAsync);
 	XGrabKey (GDK_DISPLAY (), key_code,
 			Mod2Mask,
-			GDK_ROOT_WINDOW (), True,
+			GDK_WINDOW_XID (root), True,
 			GrabModeAsync, GrabModeAsync);
 	XGrabKey (GDK_DISPLAY (), key_code,
 			Mod5Mask,
-			GDK_ROOT_WINDOW (), True,
+			GDK_WINDOW_XID (root), True,
 			GrabModeAsync, GrabModeAsync);
 	XGrabKey (GDK_DISPLAY (), key_code,
 			LockMask,
-			GDK_ROOT_WINDOW (), True,
+			GDK_WINDOW_XID (root), True,
 			GrabModeAsync, GrabModeAsync);
 	XGrabKey (GDK_DISPLAY (), key_code,
 			Mod2Mask | LockMask,
-			GDK_ROOT_WINDOW (), True,
+			GDK_WINDOW_XID (root), True,
 			GrabModeAsync, GrabModeAsync);
 	XGrabKey (GDK_DISPLAY (), key_code,
 			Mod5Mask | LockMask,
-			GDK_ROOT_WINDOW (), True,
+			GDK_WINDOW_XID (root), True,
 			GrabModeAsync, GrabModeAsync);
 	XGrabKey (GDK_DISPLAY (), key_code,
 			Mod2Mask | Mod5Mask,
-			GDK_ROOT_WINDOW (), True,
+			GDK_WINDOW_XID (root), True,
 			GrabModeAsync, GrabModeAsync);
 	XGrabKey (GDK_DISPLAY (), key_code,
 			Mod2Mask | Mod5Mask | LockMask,
-			GDK_ROOT_WINDOW (), True,
+			GDK_WINDOW_XID (root), True,
 			GrabModeAsync, GrabModeAsync);
 
 	gdk_flush ();
@@ -364,15 +386,60 @@ grab_key (int key_code)
 			 key_code);
 		acme_error (error);
 		g_free (error);
-		exit (1);
 	}
 }
 
+static void
+grab_key (Acme *acme, int key_code)
+{
+	GList *l;
+
+	for (l = acme->screens; l != NULL; l = l->next)
+	{
+		GdkScreen *screen;
+
+		screen = (GdkScreen *) l->data;
+		grab_key_real (key_code, gdk_screen_get_root_window (screen));
+	}
+}
+
+static void
+unhookup_keysym (int keycode)
+{
+	char *command;
+
+	if (keycode <= 0)
+		return;
+
+	command = g_strdup_printf ("xmodmap -e \"keycode %d = \"", keycode);
+
+	g_spawn_command_line_sync (command, NULL, NULL, NULL, NULL);
+	g_free (command);
+}
+
+static gboolean
+hookup_keysym (int keycode, const char *keysym)
+{
+	char *command;
+
+	if (keycode <= 0)
+		return TRUE;
+
+	command = g_strdup_printf ("xmodmap -e \"keycode %d = %s\"",
+			keycode, keysym);
+
+	g_spawn_command_line_sync (command, NULL, NULL, NULL, NULL);
+	g_free (command);
+
+	return FALSE;
+}
+
 static void
 update_kbd_cb (GConfClient *client, guint id, GConfEntry *entry, gpointer data)
 {
 	Acme *acme = (Acme *) data;
 	int i;
+	char *keysym = NULL;
 
 	g_return_if_fail (entry->key != NULL);
 
@@ -382,14 +449,55 @@ update_kbd_cb (GConfClient *client, guint id, GConfEntry *entry, gpointer data)
 		if (strcmp (entry->key, keys[i].key_config) == 0)
 		{
 			int key_code;
+			gboolean grab = TRUE;
+
+			ungrab_key (acme, keys[i].key_code);
+			switch (keys[i].key_type) {
+			case PLAY_KEY:
+			case PAUSE_KEY:
+			case STOP_KEY:
+			case PREVIOUS_KEY:
+			case NEXT_KEY:
+			case REFRESH_KEY:
+				unhookup_keysym (keys[i].key_code);
+				break;
+			}
+
+			keys[i].key_code = -1;
 
-			ungrab_key (keys[i].key_code);
 			key_code = gconf_client_get_int (acme->conf_client,
-					keys[i].key_config,
-					NULL);
-			if (key_code > 0)
+					keys[i].key_config, NULL);
+
+			switch (keys[i].key_type) {
+			case PLAY_KEY:
+				grab = hookup_keysym (key_code,
+						"XF86AudioPlay");
+				break;
+			case PAUSE_KEY:
+				grab = hookup_keysym (key_code,
+						"XF86AudioPause");
+				break;
+			case STOP_KEY:
+				grab = hookup_keysym (key_code,
+						"XF86AudioStop");
+				break;
+			case PREVIOUS_KEY:
+				grab = hookup_keysym (key_code,
+						"XF86AudioPrev");
+				break;
+			case NEXT_KEY:
+				grab = hookup_keysym (key_code,
+						"XF86AudioNext");
+				break;
+			case REFRESH_KEY:
+				grab = hookup_keysym (key_code,
+						"XF86Refresh");
+				break;
+			}
+
+			if (key_code > 0 && grab == TRUE)
 			{
-				grab_key (key_code);
+				grab_key (acme, key_code);
 				keys[i].key_code = key_code;
 			}
 		}
@@ -402,8 +510,29 @@ init_kbd (Acme *acme)
 	int i;
 	gboolean retval = FALSE;
 
+	acme->display = gdk_display_get_default ();
+	acme->screens = NULL;
+
+	if (gdk_display_get_n_screens (acme->display) == 1)
+	{
+		acme->screens = g_list_append (acme->screens,
+				gdk_screen_get_default ());
+	} else {
+		for (i = 0; i < gdk_display_get_n_screens (acme->display); i++)
+		{
+			GdkScreen *screen;
+
+			screen = gdk_display_get_screen (acme->display, i);
+
+			if (screen != NULL)
+				acme->screens = g_list_append (acme->screens,
+						screen);
+		}
+	}
+
 	for (i = 0; i < HANDLED_KEYS; i++)
 	{
+		gboolean grab = TRUE;
 		int tmp;
 
 		tmp = gconf_client_get_int (acme->conf_client,
@@ -412,14 +541,41 @@ init_kbd (Acme *acme)
 
 		keys[i].key_code = tmp;
 
-		if (tmp > 0)
+		switch (keys[i].key_type) {
+		case PLAY_KEY:
+			grab = hookup_keysym (keys[i].key_code,
+					"XF86AudioPlay");
+			break;
+		case PAUSE_KEY:
+			grab = hookup_keysym (keys[i].key_code,
+					"XF86AudioPause");
+			break;
+		case STOP_KEY:
+			grab = hookup_keysym (keys[i].key_code,
+					"XF86AudioStop");
+			break;
+		case PREVIOUS_KEY:
+			grab = hookup_keysym (keys[i].key_code,
+					"XF86AudioPrev");
+			break;
+		case NEXT_KEY:
+			grab = hookup_keysym (keys[i].key_code,
+					"XF86AudioNext");
+			break;
+		case REFRESH_KEY:
+			grab = hookup_keysym (keys[i].key_code,
+					"XF86Refresh");
+			break;
+		}
+
+		if (tmp > 0 && grab == TRUE)
 		{
 #ifdef DEBUG
 			g_print ("grabbed key %d for gconf key %s\n",
 					keys [i].key_code,
 					keys[i].key_config);
 #endif
-			grab_key (keys [i].key_code);
+			grab_key (acme, keys [i].key_code);
 			retval = TRUE;
 		}
 	}
@@ -460,17 +616,19 @@ about_activated (GtkMenuItem *menuitem, gpointer user_data)
 	pixbuf = gdk_pixbuf_new_from_file (ACME_DATA "acme-48.png", NULL);
 
 	about = gnome_about_new(_("Acme"), VERSION,
-			_("(C) 2001, 2002 Bastien Nocera"),
+			"Copyright \xc2\xa9 2001-2002 Bastien Nocera",
 			_("Multimedia keys daemon"),
 			(const char **)authors,
 			(const char **)documenters,
 			strcmp (translator_credits, "translator_credits") != 0 ? translator_credits : NULL,
 			pixbuf);
-
-	g_signal_connect (G_OBJECT (about), "destroy", G_CALLBACK
-			(gtk_widget_destroyed), &about);
-	g_object_add_weak_pointer (G_OBJECT (about),
-			(void**)&(about));
+	
+	if (pixbuf != NULL)
+		gdk_pixbuf_unref (pixbuf);
+	
+	g_signal_connect (G_OBJECT (about), "destroy",
+			G_CALLBACK (gtk_widget_destroyed), &about);
+	g_object_add_weak_pointer (G_OBJECT (about), (void**)&(about));
 
 	gtk_widget_show(about);
 }
@@ -573,7 +731,7 @@ init_sm (Acme *acme)
 	}
 
 	g_signal_connect (GTK_OBJECT (master), "die",
-			(GtkSignalFunc) acme_exit, acme);
+			G_CALLBACK (acme_exit), acme);
 }
 
 static gboolean
@@ -591,11 +749,15 @@ dialog_show (Acme *acme)
 	int screen_w, screen_h;
 	int x, y;
 
+	gtk_window_set_screen (GTK_WINDOW (acme->dialog), acme->current_screen);
+	gtk_widget_realize (GTK_WIDGET (acme->dialog));
+
 	gdk_window_get_geometry (GTK_WIDGET (acme->dialog)->window,
 				 &orig_x, &orig_y,
 				 &orig_w, &orig_h, &orig_d);
-	screen_w = gdk_screen_width ();
-	screen_h = gdk_screen_height ();
+
+	screen_w = gdk_screen_get_width (acme->current_screen);
+	screen_h = gdk_screen_get_height (acme->current_screen);
 
 	x = (screen_w - orig_w) / 2;
 	y = screen_h / 2 + (screen_h / 2 - orig_h) / 2;
@@ -670,7 +832,7 @@ do_shade_window_action (Acme *acme)
 }
 
 static void
-do_www_action (Acme *acme)
+do_www_action (Acme *acme, const char *url)
 {
 	char *string, *command;
 
@@ -681,7 +843,11 @@ do_www_action (Acme *acme)
 	if (string == NULL && strcmp (string, "") == 0)
 		return;
 
-	command = g_strdup_printf (string, "about:blank");
+	if (url == NULL)
+		command = g_strdup_printf (string, "about:blank");
+	else
+		command = g_strdup_printf (string, url);
+
 	execute (command, FALSE);
 
 	g_free (command);
@@ -734,6 +900,18 @@ do_eject_action (Acme *acme)
 	gtk_widget_set_sensitive (progress, TRUE);
 }
 
+static void
+do_media_action (Acme *acme)
+{
+	GnomeVFSMimeApplication *app;
+
+	app = gnome_vfs_mime_get_default_application ("audio/x-mp3");
+	if (app->requires_terminal == TRUE || app->command == NULL)
+		return;
+
+	execute (app->command, FALSE);
+}
+
 #ifdef USE_FBLEVEL
 static void
 do_brightness_action (Acme *acme, int type)
@@ -840,6 +1018,17 @@ do_action (int type, Acme *acme)
 	case EJECT_KEY:
 		do_eject_action (acme);
 		break;
+	case MEDIA_KEY:
+		do_media_action (acme);
+		break;
+	/* For GCC to be happy */
+	case PLAY_KEY:
+	case PAUSE_KEY:
+	case STOP_KEY:
+	case PREVIOUS_KEY:
+	case NEXT_KEY:
+	case REFRESH_KEY:
+		break;
 	case HOME_KEY:
 		execute ("nautilus", FALSE);
 		break;
@@ -850,7 +1039,7 @@ do_action (int type, Acme *acme)
 		execute_this_or_that ("evolution mailto:", "balsa");
 		break;
 	case SLEEP_KEY:
-		execute ("apm", TRUE);
+		execute_this_or_that ("apm", "xset dpms force off");
 		break;
 	case SCREENSAVER_KEY:
 		execute ("xscreensaver-command -lock", FALSE);
@@ -862,7 +1051,10 @@ do_action (int type, Acme *acme)
 		execute ("yelp", FALSE);
 		break;
 	case WWW_KEY:
-		do_www_action (acme);
+		do_www_action (acme, NULL);
+		break;
+	case GROUPS_KEY:
+		do_www_action (acme, "http://www.gnomedesktop.org");
 		break;
 	case CALCULATOR_KEY:
 		execute ("gnome-calculator", FALSE);
@@ -893,12 +1085,32 @@ acme_filter_events (GdkXEvent *xevent, GdkEvent *event, gpointer data)
 {
 	Acme *acme = (Acme *) data;
 	XEvent *xev = (XEvent *) xevent;
+	XAnyEvent *xanyev = (XAnyEvent *) xevent;
 	XKeyEvent *key = (XKeyEvent *) xevent;
+	GdkScreen *event_screen = NULL;
+	GList *l;
 	int i;
 
 	if (xev->type != KeyRelease)
 		return GDK_FILTER_CONTINUE;
 
+	/* Look for which screen we're receiving events */
+	for (l = acme->screens; (l != NULL) && (event_screen == NULL);
+			l = l->next)
+	{
+		GdkWindow *window;
+		GdkScreen *screen;
+
+		screen = (GdkScreen *) l->data;
+		window = gdk_screen_get_root_window (screen);
+
+		if (GDK_WINDOW_XID (window) == xanyev->window)
+		{
+			event_screen = screen;
+			break;
+		}
+	}
+
 	key = (XKeyEvent *) xevent;
 
 	for (i = 0; i < HANDLED_KEYS; i++)
@@ -907,8 +1119,21 @@ acme_filter_events (GdkXEvent *xevent, GdkEvent *event, gpointer data)
 		g_print ("comparing %d and %d\n", keys[i].key_code,
 				key->keycode);
 #endif
+
 		if (keys[i].key_code == key->keycode)
 		{
+			switch (keys[i].key_type) {
+			case PLAY_KEY:
+			case PAUSE_KEY:
+			case STOP_KEY:
+			case PREVIOUS_KEY:
+			case NEXT_KEY:
+			case REFRESH_KEY:
+				return GDK_FILTER_CONTINUE;
+			}
+
+			acme->current_screen = event_screen;
+
 			do_action (keys[i].key_type, acme);
 			return GDK_FILTER_REMOVE;
 		}
@@ -921,6 +1146,7 @@ int
 main (int argc, char *argv[])
 {
 	Acme *acme;
+	GList *l;
 
 	bindtextdomain (GETTEXT_PACKAGE, GNOMELOCALEDIR);
 	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
@@ -966,6 +1192,7 @@ main (int argc, char *argv[])
 	}
 	init_sm (acme);
 	init_tray (acme);
+	acme->current_screen = gdk_screen_get_default ();
 	acme->screen = wnck_screen_get_default ();
 	gtk_widget_realize (acme->dialog);
 	acme->dialog_timeout = 0;
@@ -995,7 +1222,17 @@ main (int argc, char *argv[])
 #endif
 
 	/* Start filtering the events */
-	gdk_window_add_filter (NULL, acme_filter_events, (gpointer) acme);
+	for (l = acme->screens; l != NULL; l = l->next)
+	{
+		GdkScreen *screen;
+		GdkWindow *window;
+
+		screen = (GdkScreen *) l->data;
+		window = gdk_screen_get_root_window (screen);
+		gdk_window_add_filter (window,
+				acme_filter_events,
+				(gpointer) acme);
+	}
 
 	gtk_main ();
 
